#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Reflection;
using System.IO;
using HeuristicLab.Persistence.Default.Xml;
using System.Diagnostics;
using HeuristicLab.Optimization;
using System.Linq;
using Microsoft.Hpc.Scheduler;
using System.ServiceModel;
using HeuristicLab.MPIAlgorithmRunner;
using HeuristicLab.Operators.MPISupport;

namespace HeuristicLab.MPIEngine {
  /// <summary>
  /// Represents an engine that executes its steps in parallel (if possible) using multiple threads.
  /// This engine is suitable for parallel processing on shared memory systems which provide multiple cores.
  /// </summary>
  [StorableClass]
  [Item("MPI Engine", "Engine for parallel execution of algorithms using multiple processes (suitable for distributed memory systems with multiple cores).")]
  public class MPIEngine : Engine {
    [StorableConstructor]
    protected MPIEngine(bool deserializing) : base(deserializing) { }
    protected MPIEngine(MPIEngine original, Cloner cloner) : base(original, cloner) { }
    public MPIEngine() : base() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MPIEngine(this, cloner);
    }

    private IAlgorithm algorithm;

    public override void Start() {
      if (ExecutionStack.Count == 1) {
        ExecutionContext context = ExecutionStack.First() as ExecutionContext;

        ExecutionContext algorithmContext = context.Parent as ExecutionContext;

        EngineAlgorithm alg = typeof(ExecutionContext).InvokeMember("parameterizedItem",
          BindingFlags.GetField | BindingFlags.NonPublic |
          BindingFlags.Instance, null, algorithmContext, null) as EngineAlgorithm;

        alg = alg.Clone() as EngineAlgorithm;
        alg.Engine = new SequentialEngine.SequentialEngine();

        algorithm = alg;
      }

      base.Start();
    }

    protected override void Run(System.Threading.CancellationToken cancellationToken) {
      if (ExecutionStack.Count == 1) {
        ExecutionContext context = ExecutionStack.Pop() as ExecutionContext;

        IScope globalScope = context.Scope;

        string exec = @"mpiexec";
        string args = @"-c 3 C:\public\MPISupport\HeuristicLab.MPIAlgorithmRunner-3.3.exe ";

        IScheduler scheduler = new Scheduler();
        scheduler.Connect("blade00.hpc.fh-hagenberg.at");

        ISchedulerJob job = scheduler.CreateJob();
        job.Name = "HeuristicLab.MPIEngine";
        job.RequestedNodes.Add("BLADE03");
        ISchedulerTask task = job.CreateTask();
        task.Name = "HeuristicLab.MPIAlgorithmRunner";
        task.CommandLine = exec + " " + args;
        task.StdOutFilePath = "stdout.txt";
        task.StdErrFilePath = "stderr.txt";
        task.MinimumNumberOfCores = task.MaximumNumberOfCores = 3;
        job.AddTask(task);

        scheduler.SubmitJob(job, @"HPC\svonolfe", @"Vunkopf!3");

        try {
          string address = null;
          int timeout = 10;

          while (address == null && timeout > 0) {
            ISchedulerJob schedulerJob = scheduler.OpenJob(job.Id);
            if (schedulerJob != null) {
              INameValueCollection properties = schedulerJob.GetCustomProperties();
              NameValue item = properties.FirstOrDefault(i => i.Name == "address");
              if (item != null) {
                address = item.Value;
              } else {
                System.Threading.Thread.Sleep(1000);
                timeout--;
              }
            }
          }

          if (address == null) {
            throw new Exception("A timeout occurred when starting the MPIAlgorithmRunner");
          }

          NetTcpBinding netTCPBinding = new NetTcpBinding(SecurityMode.None);
          ChannelFactory<IAlgorithmBroker> factory = new ChannelFactory<IAlgorithmBroker>(netTCPBinding, address);
          IAlgorithmBroker proxy = factory.CreateChannel();

          proxy.TransmitAlgorithm(new MPITransportWrapper<IAlgorithm>(algorithm));

          while (!proxy.IsAlgorithmTerminated()) {
            ItemList<ResultCollection> results = proxy.GetResults().InnerItem;

            ResultCollection resultCollection = (globalScope.Variables["Results"].Value as ResultCollection);

            if (resultCollection != null && results != null) {
              if (!resultCollection.ContainsKey("MPIResults"))
                resultCollection.Add(new Result("MPIResults", results));

              resultCollection["MPIResults"].Value = results;
            }

            System.Threading.Thread.Sleep(5000);
          }
        }
        catch (Exception e) {
          scheduler.CancelJob(job.Id, "Exception: " + e.Message);
          throw e;
        }
      }
    }
  }
}
